import datetime
import os
import socket


class Status:
    def __init__(self, dir_):
        self.dir = dir_
        self.status = "status"
        for status in ["Done", "Failed", "Running"]:
            if os.path.exists(os.path.join(dir_, status)):
                self.status = status

        self.status_path = os.path.join(dir_, self.status)

    def done(self):
        self._set_status("Done")

    def running(self):
        self._set_status("Running")

    def failed(self):
        self._set_status("Failed")

    def _set_status(self, status):
        timestamp = datetime.datetime.now().strftime("%y/%m/%d-%H:%M:%S")

        if not os.path.exists(self.status_path):
            with open(self.status_path, "w") as f:
                f.write("Host: " + socket.gethostname() + "\n")

        with open(self.status_path, "a") as status_file:
            status_file.write(status + ": " + timestamp + "\n")

        status_path = os.path.join(self.dir, status)
        os.rename(self.status_path, status_path)

        self.status = status
        self.status_path = status_path



from pype import cli


def test_pipeline_can_run(pipeline_config, pipeline_dir):
    cli.run_(pipeline_config, log=False, tag=None)


def test_pipeline_only_runs_tags(pipeline_config, pipeline_dir):
    cli.run_(pipeline_config, log=False, tag='a')



from pype.baseconfig import BaseConfig


class Config(BaseConfig):
    script_path = "example/job.py"
    params = {"append": "good day"}
    outputs = {"msg": "msg.txt"}


def main(config, logger):
    logger.info('running job')



import atexit
import logging
import os
import subprocess

import __main__
import yaml

GIT_CONTROL = os.system("git rev-parse") == 0


def get_git_sha():
    return (
        subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
        .decode("ascii")
        .strip()
    )


def save_git_sha(job_dir):
    cmd = f"echo {get_git_sha()} > {job_dir}/git_sha.txt"
    os.system(cmd)


def get_pipeline_dir():
    pipeline_dir = os.path.relpath(os.path.dirname(__main__.__file__))
    if not os.path.exists(pipeline_dir):
        os.makedirs(pipeline_dir, exist_ok=True)
        logging.info(f"Created pipeline directory {pipeline_dir}")

    return pipeline_dir


class ConfigCollector:
    _instance = None

    def __init__(self):
        raise Exception(
            "This class is a singleton. Use JobCollector.instance instead of __init__"
        )

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls._init(cls._instance)
        return cls._instance

    # pylint:disable=attribute-defined-outside-init
    def _init(self):
        self.configs = []
        self.pipeline_dir = get_pipeline_dir()
        atexit.register(self.run)

    def add_config(self, config):
        self.configs.append(config)

    def run(self):
        yaml.dump(
            self.configs,
            open(os.path.join(self.pipeline_dir, "pipeline_config.yaml"), "w"),
        )


CONFIG_COLLECTOR = ConfigCollector.instance()



import logging


def get_logger(job_id, file_path):
    logger = logging.getLogger(job_id)
    logger.setLevel(logging.INFO)

    # Create handlers
    c_handler = logging.StreamHandler()
    f_handler = logging.FileHandler(file_path, 'w')
    c_handler.setLevel(logging.INFO)
    f_handler.setLevel(logging.INFO)

    # Create formatters and add it to handlers
    c_format = logging.Formatter('%(name)s - %(asctime)s - %(message)s')
    f_format = logging.Formatter('%(name)s - %(asctime)s - %(message)s')
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)

    # Add handlers to the logger
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)

    return logger



# pylint: disable=no-member
import os
import pprint
from uuid import uuid4

import yaml

from pype import utils


class BaseConfig:
    _expected_attribute_types = {
        "inputs": set,
        "optional_inputs": set,
        "script_path": str,
        "params": dict,
        "outputs": dict,
    }

    def __init__(
        self,
        pipeline_dir=None,
        job_id=None,
        inputs=None,
        params=None,
        wait_for=None,
        tag=None,
    ):
        """
        Inherit from this config class to create configs for jobs.

        class ExampleConfig(BaseConfig):
            script_path = 'path/to/script/that/configured/by/this/config'
            inputs = {'input_key'}
            params = {'param_key'}
            outputs = {'output_1': 'path/to/where_output_is_saved'}
            optional_inputs = {'optional_input'}

        script pth has to be set so that the config knows where to find the

        To set an optional parameter/input use the default value None.
            outputs has to be a dictionary as it is name of what to save and where to save it
            inputs and params are sets as the user has to specify themselves what there values re
            inputs are always paths to data or outputs from other jobs
            params are the parameters needed to run a job and does not necessarily come from
            outside

        To set an optional parameter/input use the default value None.
        """

        # Verify construcion of child
        assert hasattr(
            self, "script_path"
        ), f"Configuration class for f{type(self)} must include script_path"

        for attr_name in self._expected_attribute_types:
            if not hasattr(self, attr_name):
                setattr(self, attr_name, self._expected_attribute_types[attr_name]())

        self._verify_attributes()

        # Initialise config

        inputs = inputs if inputs else dict()
        params = params if params else dict()

        _verify_keys(inputs.keys(), self.inputs, self.optional_inputs)
        _verify_keys(params.keys(), set(), self.params.keys())

        job_id = str(uuid4()) if not job_id else job_id
        if not pipeline_dir:
            pipeline_dir = utils.get_pipeline_dir()
        job_dir = os.path.join(pipeline_dir, 'jobs', job_id)
        output_dir = os.path.join(job_dir, "output")

        outputs = dict()
        for key in self.outputs:
            outputs[key] = os.path.join(output_dir, self.outputs[key])

        self.config = {}
        self.config["script_path"] = self.script_path
        self.config["job_id"] = job_id
        self.config["job_dir"] = job_dir
        self.config["qualname"] = type(self).__qualname__
        if wait_for:
            self.config["wait_for"] = wait_for
        if tag:
            if isinstance(tag, list):
                self.config["tag"] = tag
            else:
                self.config["tag"] = [tag]

        for key in self.params:
            if key not in params:
                params[key] = self.params[key]

        for key in self.optional_inputs:
            if key not in inputs:
                inputs[key] = None

        for name, val in zip(
            ["params", "inputs", "outputs"], [params, inputs, outputs]
        ):
            if val:
                self.config[name] = val

        os.makedirs(output_dir, exist_ok=True)
        print(f"Created dirs for {job_id}: {os.path.join(job_dir, 'config.yaml')}")

        yaml.dump(self.config, open(os.path.join(job_dir, "config.yaml"), "w"))

        if utils.GIT_CONTROL:
            utils.save_git_sha(job_dir)

        utils.CONFIG_COLLECTOR.add_config(self.config)

    def _verify_attributes(self):
        attrs = set(filter(lambda x: not x.startswith("_"), dir(self)))
        unexpected_attributes = attrs - self._expected_attribute_types.keys()
        if unexpected_attributes:
            raise AttributeError(
                f"Got unexpected attributes {unexpected_attributes}, allowed "
                f"attributes are {self._expected_attribute_types}"
            )

        for attr in attrs:
            if attr in self._expected_attribute_types:
                assert isinstance(
                    getattr(self, attr), self._expected_attribute_types[attr]
                ), (
                    f"Attribute {attr} on {type(self)} should be of type "
                    f"{self._expected_attribute_types[attr]} but is instead of type"
                    f" {type(getattr(self,attr))}."
                )

    def __getitem__(self, key):
        return self.config[key]

    def __str__(self):
        return pprint.pformat(self.__dict__, indent=3)


def _verify_keys(data_keys, expected_keys, optional_keys):
    missing_keys = expected_keys - data_keys
    unexpected_keys = data_keys - expected_keys - optional_keys

    if missing_keys:
        raise RuntimeError(f"missing inputs/params: {missing_keys}")
    if unexpected_keys:
        raise RuntimeError(f"Got unexpected_keys inputs/params: {unexpected_keys}")



import csv
import os

from pype import utils


class ResultsCollector():
    def __init__(self, fieldnames):
        pipeline_dir = utils.get_pipeline_dir()

        self.path = os.path.join(pipeline_dir, 'results.csv')
        self.fieldnames=fieldnames
        with open(self.path, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()

    def write(self, results):
        with open(self.path, 'a') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=self.fieldnames)
            writer.writerow(results)



import os

import pytest
import yaml

from pype import BaseConfig

params = {"param": 1}
inputs = {"input": 1, "optional_input": 2}


def test_mockconfig_can_instantiate(tmpdir, MockConfig,):
    MockConfig(job_id="bla", pipeline_dir=tmpdir, params=params, inputs=inputs)


def test_mockconfig_can_has_attributes(tmpdir, MockConfig,):
    config = MockConfig(job_id="bla", pipeline_dir=tmpdir, params=params, inputs=inputs)

    for key in ["inputs", "params", "outputs", "job_id"]:
        assert key in config.config


def test_mockconfig_can_instantiate_twice(tmpdir, MockConfig,):
    MockConfig(job_id="foo", pipeline_dir=tmpdir, params=params, inputs=inputs)
    config = MockConfig(job_id="bar", pipeline_dir=tmpdir, params=params, inputs=inputs)

    assert config["outputs"]["result"].endswith("bar/output/result.ext")


def test_mock_config_creates_the_right_paths(tmpdir, MockConfig,):
    config = MockConfig(job_id="bla", pipeline_dir=tmpdir, params=params, inputs=inputs)

    assert os.path.exists(os.path.join(tmpdir, config["job_id"], "output"))
    assert os.path.exists(os.path.join(tmpdir, config["job_id"], "git_sha.txt"))
    assert os.path.exists(os.path.join(tmpdir, config["job_id"], "config.yaml"))


def test_cant_instantiate_when_missing_input(tmpdir, MockConfig,):
    with pytest.raises(RuntimeError):
        MockConfig(params=params, pipeline_dir=tmpdir, inputs=dict())


def test_cant_instantiate_with_unexpected_input(tmpdir, MockConfig,):
    with pytest.raises(RuntimeError):
        inputs_unexpected = dict({"unexpected_input": "2"}, **inputs)
        MockConfig(params=params, inputs=inputs_unexpected, pipeline_dir=tmpdir)


def test_cant_instantiate_with_unexpected_param(tmpdir, MockConfig,):
    with pytest.raises(RuntimeError):
        params_unexpected = dict({"unexpected_param": "2"}, **params)
        MockConfig(params=params_unexpected, inputs=inputs, pipeline_dir=tmpdir)


def test_cant_create_class_with_wrong_attributes(tmpdir, MockConfig,):
    class WrongConfig(BaseConfig):
        script_path = "script/path"
        wrong_name_for_inputs = {"hello"}

    with pytest.raises(AttributeError):
        WrongConfig(pipeline_dir=tmpdir)


def test_mockconfig_has_default_value_on_param(tmpdir, MockConfig,):
    config = MockConfig(job_id="bla", pipeline_dir=tmpdir, inputs=inputs)
    assert config['params']['param'] == 'default'

def test_mockconfig_has_none_on_optional_input(tmpdir, MockConfig,):
    config = MockConfig(job_id="bla", pipeline_dir=tmpdir, inputs={'input': 1})
    assert config['inputs']['optional_input'] is None

def test_no_params_config_can_instantiate(tmpdir, MockConfig,):
    class NoParamConfig(BaseConfig):
        script_path = 'test.py'
        inputs = {"msg"}
        outputs = {"processed": "processed_message.txt"}

    NoParamConfig(
        pipeline_dir=tmpdir,
        job_id = "test_job",
        inputs={"msg": "msg.txt"}
    )

def test_inheritance():
    class ConfigMixin():
        outputs = {"a": "A"}

    class Config(ConfigMixin, BaseConfig):
        script_path ="b"

    config = Config()
    assert config['outputs']



import os

import pytest

from pype import BaseConfig
from pype import utils


@pytest.fixture
def MockConfig():
    class _MockConfig(BaseConfig):
        script_path = "/mock/path"
        params = {"param": "default"}
        inputs = {"input"}
        optional_inputs = {"optional_input"}
        outputs = {"result": "result.ext"}

    return _MockConfig


@pytest.fixture
def pipeline_dir(tmpdir, MockConfig):  # pylint: disable=redefined-outer-name
    MockConfig(job_id="A1", pipeline_dir=tmpdir, inputs={"input": 1}, tag="a")
    MockConfig(job_id="A2", pipeline_dir=tmpdir, inputs={"input": 1}, tag="a")
    MockConfig(job_id="B1", pipeline_dir=tmpdir, inputs={"input": 1}, tag="b")
    MockConfig(job_id="B2", pipeline_dir=tmpdir, inputs={"input": 1}, tag="b")
    utils.CONFIG_COLLECTOR.run(tmpdir)
    return tmpdir


@pytest.fixture
def pipeline_config(pipeline_dir): # pylint: disable=redefined-outer-name
    return os.path.join(pipeline_dir, 'pipeline_config.yaml')



from setuptools import find_packages, setup

setup(
    author="Mathias Schreiner",
    author_email="matschreiner@gmail.com",
    url="https://gitlab.com/matschreiner/pype",
    download_url="https://gitlab.com/matschreiner/pype/-/archive/v2.3.0/pype-v2.3.0.tar.gz",
    name="pype-ms",
    version="2.3.0",
    packages=find_packages(),
    install_requires=[
        "pyyaml",
        "pytest",
        "Click",
    ],
    entry_points={
        "console_scripts": [
            "pype = pype.cli:cli",
        ],
    },
)



import importlib
import inspect
import os
import pprint
import shutil
import socket

import click
import yaml
from pype import utils
from pype.log import get_logger
from pype.status import Status


@click.group()
def cli():
    pass


@cli.command()
@click.option("--tag", "-t", default="", help="only run configs with the tag")
@click.option("--force", "-f", default=False, is_flag=True)
@click.option("--scratch", "-f", default=None)
@click.argument("config")
def run(config, tag, force, scratch):
    run_(config, tag, force=force, scratch=scratch)


def run_(config, tag, force=False, scratch=None):
    if isinstance(config, str):
        assert config.endswith("yaml"), "Configuration file must be YAML"
        config = yaml.load(open(config, "r"), Loader=yaml.FullLoader)

    if isinstance(config, list):
        for config_ in config:
            run_(config_, tag, force=force, scratch=scratch)

    else:
        status = Status(config["job_dir"])
        if tag:
            if not tag in config.get("tag", ""):
                return

        if force:
            pass

        elif status.status == "Running" or status.status == "Done":
            print(
                f"{config['job_id']} was skipped because status is {status.status} ... "
            )
            return


        output_dir = os.path.join(config["job_dir"], "output")
        if os.path.islink(output_dir):
            os.remove(output_dir)
            os.makedirs(output_dir)

        if scratch:
            scratch_output_dir = os.path.join(scratch, config["job_id"])
            os.makedirs(scratch_output_dir, exist_ok=True)
            shutil.rmtree(output_dir)
            os.symlink(scratch_output_dir, output_dir)

        print(f"{config['job_id']} is running")
        run_job(config)

        if scratch:
            os.remove(output_dir)
            shutil.copytree(scratch_output_dir, output_dir)
            shutil.rmtree(scratch_output_dir)


def run_job(config):
    job_dir = config["job_dir"]
    log_file = os.path.join(job_dir, "file.log")
    logger = get_logger(config["job_id"], log_file)
    status = Status(job_dir)

    msg = running_job_msg(config)
    logger.info(msg)
    utils.save_git_sha(job_dir)

    status.running()

    try:
        module = _import_module(config["script_path"])
        if not hasattr(module, "main"):
            raise RuntimeError(f"{config['script_path']} has no main function.")

        if (
            "logger" in inspect.getargspec(module.main).args
        ):  # pylint: disable=deprecated-method
            module.main(config, logger)
        else:
            module.main(config)
        status.done()
        logger.info("job terminated succesfully.\n\n-\n")

    except Exception:  # pylint: disable=broad-except
        status.failed()
        logger.exception("Exception occurred: \n\n")


def running_job_msg(config):
    space = 4 * " "

    msg = f"Running job {config['job_id']}"
    hashs = (2 * len(space) + len(msg)) * "#" + ""

    full_msg = f"\n{hashs}\n{'    '+msg}\n{hashs}\n\n"
    full_msg += "Host: " + socket.gethostname() + "\n"
    full_msg += "PID: " + str(os.getpid()) + "\n"
    full_msg += "gitsha: " + str(utils.get_git_sha()) + 2 * "\n"
    full_msg += "Configuration:\n"

    full_msg += pprint.pformat(config)
    length = max([len(l) for l in pprint.pformat(config).split("\n")])
    full_msg += 2 * "\n" + length * "_" + "\n"

    return full_msg


def permission_to_continue(msg):
    return input(msg + "Type 'y' or 'yes' to continue anyways\n").lower() in [
        "y",
        "yes",
    ]


def _import_module(path):
    spec = importlib.util.spec_from_file_location("module.name", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def _uncomitted():
    if not utils.GIT_CONTROL:
        return False

    cmd = r"git status | grep -q '\smodified:\s'"
    code = os.system(cmd)
    return code == 0
